# Copyright 2005 PyTom <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import cStringIO
import Image
import ImageDraw
import gtk
import gtk.gdk
import gobject
import time
import sys

import alice
import _alice



def load_spng(image):
    
    f = image.path.open()

    format = f.readline().strip()
    if format != "ALICE-SPNG-1":
        raise Exception("Invalid SPNG format.")

    width = int(f.readline())
    height = int(f.readline())
    xo = int(f.readline())
    yo = int(f.readline())

    sf = cStringIO.StringIO(f.read())

    f.close()
    del f

    im = Image.open(sf)

    r, g, b, a = im.split()
    im = im.convert("RGBX")

    rv = (width, height), (xo, yo), im, a

    return rv


cache = { }
cache_times = { }
cache_time = 0

# TODO: Need to add caching (and cache invalidation) to this.
def load_image(image, state):

    recolor = state.recolors.get((image.layer.category, image.layer.name, image.name), None)

    cache_times[image, recolor] = cache_time

    if (image, recolor) in cache:
        return cache[image, recolor]

    (width, height), (xo, yo), im, a = load_spng(image)

    if recolor:
        _alice.recolor(im.getim(), 
                       recolor.h_offset,
                       recolor.s_multiply,
                       recolor.s_offset,
                       recolor.l_multiply,
                       recolor.l_offset)

    rv = (width, height), (xo, yo), im, a

    cache[image, recolor] = rv
    return rv

def purge_cache():

    global cache_time

    for k in cache.keys():
        if cache_times[k] < cache_time - 2:
            del cache[k]
            del cache_times[k]

    cache_time += 1


def render(state, config, scaling=Image.NEAREST, grid=False):

    global cache
    global newcache

    # Figure out the doll.
    doll = alice.dolls.dolls[state.doll]

    # Find active images
    images = [ ]

    for cat, layer in state.configs[config]:

        l = doll.layers.get(cat, { }).get(layer, None)

        if l is None:
            continue

        for i in l.images:
            images.append(i)


    images = [ (-i.priority, i) for i in images ]
    images.sort()

    # Render a stack of images.
    rv = None

    for negprio, i in images:
        size, off, im, alpha = load_image(i, state)

        if not rv:
            rv = Image.new("RGBA", size)
        
        rv.paste(im, off, alpha)

    if not rv:
        return None

    scale = (state.h.scale, state.v.scale)
    if rv.size != scale:
        rv = rv.resize(scale, scaling)

    (cw, ch) = state.h.crop, state.v.crop
    (cx, cy) = state.h.offset, state.v.offset

    if rv.size != (cw, ch):
        rv = rv.crop((cx, cy, cx+cw, cy+ch))

    purge_cache()

    return rv

class RenderWindow(object):

    def __init__(self):
        self.state = None
        self.config = None
        self.queued = False

        self.window = gtk.Window()
        self.window.set_title("Render Preview")

        sw = gtk.ScrolledWindow()
        self.window.add(sw)

        self.image = gtk.Image()
        self.viewport = gtk.Viewport()
        self.viewport.add(self.image)

        sw.add(self.viewport)
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)

        self.window.connect("delete-event", self.destroy)

        self.queued_at = None

        self.old_size = None


        # Cache of background grid.
        self.grid = None
        self.grid_size = None

        gobject.timeout_add(250, self.render)

    def destroy(self, w, e):
        w.hide()
        return True

    def queue_render(self, state, config):

        self.state = state
        self.config = config
        self.queued_at = time.time()

    def draw_grid_under(self, im):

        if self.grid and self.grid_size == im.size:
            rv = self.grid.copy()
            rv.paste(im, None, im)
            return rv

        rv = Image.new("RGB", im.size, (0x66, 0x66, 0x66))
        draw = ImageDraw.Draw(rv)

        width, height = im.size

        edge = 16

        y = 0
        ysquare = 0

        while y < height:

            x = 0
            xsquare = 0

            while x < width:

                if (xsquare + ysquare) & 1:
                    draw.rectangle([ (x, y), (x + edge - 1, y + edge - 1) ],
                                   fill=(0x99, 0x99, 0x99))

                xsquare += 1
                x += edge

            ysquare += 1
            y += edge


        self.grid = rv
        self.grid_size = rv.size

        rv = rv.copy()
        rv.paste(im, None, im)

        return rv

    def render(self):

        try:
            try:


                if not self.queued_at:
                    return True

                # Only begin a render if one's been queued for .25 seconds.
                if self.queued_at > time.time() - .25:
                    return True

                self.queued_at = None

                if not self.state or not self.config:
                    return True

                im = render(self.state, self.config)

                if not im:
                    self.image.set_from_pixbuf(None)
                    return True

                im = self.draw_grid_under(im)

                w, h = im.size

                pb = gtk.gdk.pixbuf_new_from_data(im.tostring(),
                                                  gtk.gdk.COLORSPACE_RGB,
                                                  False,
                                                  8,
                                                  w,
                                                  h,
                                                  3 * w )

                self.image.set_from_pixbuf(pb)

                if (w, h) != self.old_size:

                    self.old_size = (w, h)

                    winw = min(gtk.gdk.screen_width() - 64, w)
                    winh = min(gtk.gdk.screen_height() - 64, h) 

                    self.image.set_size_request(w, h)
                    self.window.resize(winw + 4, winh + 4)

                self.window.show_all()

            except:
                import traceback
                traceback.print_exc()
                
        finally:
            return True

        return True

def render_all(state, dir):


    try:
        for i in state.configs:
            im = render(state, i, scaling=Image.ANTIALIAS)

            if im is None:
                continue

            im.save(dir + "/" + i + ".png", "PNG", optimize=9)

        alice.interface.messagebox(None, "Rendering complete.", False)

    except:
        type, e, tb = sys.exc_info()
        alice.interface.messagebox(None, str(e), True)
        raise
        

        
    
